#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of svgis.
# https://github.com/fitnr/svgis

# Licensed under the GNU General Public License v3 (GPLv3) license:
# http://opensource.org/licenses/GPL-3.0
# Copyright (c) 2016, Neil Freeman <contact@fakeisthenewreal.org>
import logging

'''
Utilities for manipulating the DOM and applying styles to same.
'''

SVG_NS = 'http://www.w3.org/2000/svg'


def cdata(text):
    '''Wrap text in CDATA section.'''
    return '<![CDATA[' + text + ']]>'


def ns(tag):
    '''Apply the SVG namespace to a tag, e.g. ``g`` to ``{http://www.w3.org/2000/svg}g``.'''
    return '{' + SVG_NS + '}' + tag


def apply_rule(doc, rule):
    '''
    Apply a tinycss Rule to an ElementTree.Element
    (only tested on documents created by SVGIS).
    '''
    tokenlist = _build_tokenlist(rule.selector)

    declaration = u' '.join(
        u'{}:{};'.format(d.name, d.value.as_css()) for d in rule.declarations
    )

    for tokens in tokenlist:
        # Starts as None as a marker to look into document,
        # instead of filtering the list.
        els = None
        while len(tokens) > 0:
            els, tokens = _process_tokens(doc, els, tokens)

        if els:
            for el in els:
                el.attrib['style'] = el.attrib.get('style', u'') + declaration


def _match_classes(elem_classes, rule_classes):
    '''Check if rule_classes all fall in elem_classes.'''
    return all([c in elem_classes for c in rule_classes])


def _idrule(idname):
    # Get rid of the hash here.
    return "*[@id='{}']".format(idname[1:])


def _build_tokenlist(tokens):
    tokenlist = [[]]
    itertokens = iter(tokens)

    for token in itertokens:
        if token.is_container:
            tokenlist[-1].append(token)
            continue

        if token.value == ',':
            tokenlist.append([])
            continue

        tokenlist[-1].append(token)

    return tokenlist


def _find_classes(doc, els, classes):
    '''Get elements in doc that match given classes.'''
    if els is None:
        els = doc.iter()

    return [e for e in els if _match_classes(e.attrib.get('class', []), classes)]


def _get_elems_by_tagname(doc, els, tagname):
    '''Find elems in els (or doc) that have the given tagname.'''
    if els is None:
        return doc.findall('.//' + tagname)
    else:
        return [e for e in els if e.tag == tagname]


def _unsupported_token(token):
    return token.is_container or token.type == 'ATKEYWORD' or token.value in ('+', '~', ':')


def _process_tokens(doc, els, tokens):
    '''
    Work through the given selector tokens, attempting to return the subset
    of ``els`` that match.

    Args:
        doc (ElementTree.Element): Parent element to search within.
        els (list): Current list of elements. For the first pass, this is None.
        tokens (list): List of ``tinycss.Token``s.
    '''
    if any(_unsupported_token(t) for t in tokens):
        return [], []

    if tokens[0].value == '*':
        if els is None:
            els = doc.findall('.//')

        remaining_tokens = tokens[1:]

    # Look inside the given elements!
    elif tokens[0].type == 'S':
        try:
            els = [u for e in els for u in e.findall('.//')]
            remaining_tokens = tokens[1:]

        except TypeError:
            els = None
            remaining_tokens = tokens[1:]

    # Child operator is easy.
    elif tokens[0].value == '>' and tokens[0].type == 'DELIM':
        remaining_tokens = tokens[1:]
        els = [u for e in els for u in e.findall('./')]

    # Is there a <HASH>: it's an ID, find that, then look for other classes or tags.
    elif any(t.type == 'HASH' for t in tokens):
        idtokens = [t for t in tokens if t.type == 'HASH']

        # Remove all the id hashes.
        remaining_tokens = [t for t in tokens if t not in idtokens]

        # Allow for multiple id hashes, but we only want the first one
        els = doc.findall('.//' + _idrule(idtokens[0].value))

    # first token is <DELIM '.'>: make a list of classes
    elif tokens[0].value == '.' and tokens[0].type == 'DELIM':
        classes = []

        for tok in tokens:
            if tok.value == '.' and tok.type == 'DELIM':
                continue

            elif tok.type == 'IDENT':
                classes.append(tok)

            else:
                break

        try:
            els = _find_classes(doc, els, [c.value for c in classes])
            # All tokens after the last class in classes.
            remaining_tokens = tokens[tokens.index(classes[-1]) + 1:]
        except ValueError:
            remaining_tokens = tokens[1:]

    # first token is <IDENT>: it's an element, possibly followed other stuff.
    elif tokens[0].type == 'IDENT':
        els = _get_elems_by_tagname(doc, els, ns(tokens[0].value))
        remaining_tokens = tokens[1:]

    # Didn't recognize that token!
    else:
        logging.getLogger('svgis').warn('Unknown CSS: %s', tokens[0].value)
        remaining_tokens = tokens[1:]

    return els, remaining_tokens



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of svgis.
# https://github.com/fitnr/svgis

# Licensed under the GNU General Public License v3 (GPLv3) license:
# http://opensource.org/licenses/GPL-3.0
# Copyright (c) 2016, Neil Freeman <contact@fakeisthenewreal.org>

import unittest
from svgis import projection

SHP = 'tests/test_data/cb_2014_us_nation_20m.shp'


class ProjectionTestCase(unittest.TestCase):

    def testUtm(self):
        assert projection.utm_proj4(-21, 42) == '+proj=utm +zone=27 +north +datum=WGS84 +units=m +no_defs'

        assert projection.utm_proj4(-21, -42) == '+proj=utm +zone=27 +south +datum=WGS84 +units=m +no_defs'

        with self.assertRaises(ValueError):
            projection.utm_proj4(-200, 100)

    def testlayer_bounds(self):
        assert projection.layer_bounds(SHP) == (-179.174265, 17.913769, 179.773922, 71.352561)
        fix = (-9183694.888482913, -2282590.8349678256, 9343945.82964409, 10263876.62548214)
        for a in zip(fix, projection.layer_bounds(SHP, 'EPSG:102009')):
            self.assertAlmostEqual(*a)

    def testLocalTm(self):
        fixture = ('+proj=lcc +lon_0=0 +lat_1=0 +lat_2=0 +lat_0=0'
                   '+x_0=0 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0'
                   '+units=m +no_defs')
        self.assertEqual(projection.tm_proj4(0, 0, 0), fixture)

    def testTransformBounds(self):
        bounds = (-74, 42, -73, 43)

        with self.assertRaises(TypeError):
            projection.transform_bounds(None, {'init': 'epsg:4269'}, bounds)

        with self.assertRaises(TypeError):
            projection.transform_bounds({'init': 'epsg:4269'}, None, bounds)

        a = projection.transform_bounds({'init': 'epsg:4269'}, {'init': 'epsg:3102'}, bounds)

        fixture = (43332273.50269379, 15584115.894447982, 44004519.424246654, 16320640.928220816)

        for z in zip(a, fixture):
            self.assertAlmostEqual(*z)

    def testGenerateCRS(self):
        bounds = -82.2, 40.1, -78.9, 45.8
        a = projection.generatecrs(*bounds, proj_method='utm')
        self.assertEqual(a, '+proj=utm +zone=17 +north +datum=WGS84 +units=m +no_defs')

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of svgis.
# https://github.com/fitnr/svgis

# Licensed under the GNU General Public License v3 (GPLv3) license:
# http://opensource.org/licenses/GPL-3.0
# Copyright (c) 2016, Neil Freeman <contact@fakeisthenewreal.org>
import unittest
from xml.dom import minidom
from svgis import style, svg

try:
    basestring
except NameError:
    basestring = str


class SvgTestCase(unittest.TestCase):

    def setUp(self):
        self.file = 'tests/test_data/test.svg'
        self.newstyle = 'stroke {color:red;}'

    def test_rescale(self):
        new = style.rescale(self.file, 100)

        g = minidom.parseString(new).getElementsByTagName('g')[0]
        assert 'scale(100)' in g.attributes.get('transform').value

    def test_create(self):
        s = svg.drawing((100, 100), [])
        assert isinstance(s, basestring)

        s = svg.drawing((100, 100), [])
        assert isinstance(s, basestring)

        s = svg.drawing((100, 100), [], style=self.newstyle)

        assert self.newstyle in s

    def testGroup(self):
        g = svg.group()
        self.assertIsInstance(g, basestring)

        g = svg.group(transform="translate(10,10)")
        self.assertIn('transform="translate(10,10)"', g)

        g = svg.group(transform="scale(10)")
        assert 'transform="scale(10)"' in g

    def testAttribs(self):
        args = {
            'transform': 'translate(10, 10)',
            'fill': 'black'
        }
        assert 'transform="translate(10, 10)"' in svg.toattribs(**args)
        assert 'fill="black"' in svg.toattribs(**args)

    def testDrawCircle(self):
        point = svg.circle((0.0, 0.0), r=2)
        assert isinstance(point, basestring)
        assert 'r="2"' in point
        self.assertIn('cy="0.0"', point)
        assert 'cx="0.0"' in point

    def testDrawPath(self):
        coordinates = [
            (0.0, 0.0), (10.0, 0.0), (10.0, 10.0), (0.0, 10.0), (0.0, 0.0), 'M',
            (4.0, 4.0), (4.0, 5.0), (5.0, 5.0), (5.0, 4.0), (4.0, 4.0), 'Z'
        ]

        path = svg.path(coordinates)
        assert isinstance(path, basestring)
        assert ' M ' in path
        assert 'Z"' in path
        assert '10.0,0.0' in path
        

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of svgis.
# https://github.com/fitnr/svgis

# Licensed under the GNU General Public License v3 (GPLv3) license:
# http://www.opensource.org/licenses/GNU General Public License v3 (GPLv3)-license
# Copyright (c) 2016, Neil Freeman <contact@fakeisthenewreal.org>

try:
    from svgis import svgis
except ImportError:
    from build.lib.svgis import svgis

if __name__ == '__main__':
    shp = 'tests/test_data/cb_2014_us_nation_20m.shp'
    css = 'polygon{fill:green}'
    PROJECTION = '+proj=lcc +lat_1=20 +lat_2=60 +lat_0=40 +lon_0=-96 +x_0=0 +y_0=0 +datum=NAD83 +units=m +no_defs'
    BOUNDS = (-124.0, 20.5, -64.0, 49.0)

    svgis.map(shp, style=css, scale=1000, crs=PROJECTION, bounds=BOUNDS, clip=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of svgis.
# https://github.com/fitnr/svgis

# Licensed under the GNU General Public License v3 (GPLv3) license:
# http://www.opensource.org/licenses/GNU General Public License v3 (GPLv3)-license
# Copyright (c) 2016, Neil Freeman <contact@fakeisthenewreal.org>

import sys
from signal import signal, SIGPIPE, SIG_DFL
import logging
import click
from . import projection
from . import style as _style, svgis, __version__


none = {
    'flag_value': None,
    'expose_value': False,
    'help': '(not enabled)'
}

try:
    import shapely
    clipkwargs = {
        'default': True,
        'flag_value': True,
        'help': "Clip shapes to bounds. Slightly slower, produces smaller files (default: clip)."
    }
except ImportError:
    clipkwargs = none

try:
    import visvalingamwyatt
    simplifykwargs = {
        'type': click.IntRange(1, 100, clamp=True),
        'metavar': 'FACTOR',
        'help': ('Simplify geometries, '
                 'accepts an integer between 1 and 100, '
                 'the percentage of each geometry to retain.'),
    }
except ImportError:
    simplifykwargs = none

CLICKARGS = {
    'context_settings': dict(help_option_names=['-h', '--help'])
}

csskwargs = {
    'flag_value': True,
    'default': False,
    'help': ('Inline CSS styles to each element. '
             'Slightly slower, but required by some clients (e.g. Adobe) '
             '(default: do not inline).'),
}

inp = click.argument('layer', default=sys.stdin, type=click.File('rb'))
outp = click.argument('output', default=sys.stdout, type=click.File('wb'))


# Base
@click.group(**CLICKARGS)
@click.version_option(version=__version__, message='%(prog)s %(version)s')
@click.pass_context
def main(context):
    context.log = logging.getLogger('svgis')
    context.log.setLevel(logging.WARN)
    ch = logging.StreamHandler()
    ch.setLevel(logging.WARN)
    context.log.addHandler(ch)

    signal(SIGPIPE, SIG_DFL)


# Style
style_help = ("Style to append to SVG. "
              "Either a valid CSS string, a file path (must end in '.css'). "
              "Use '-' for stdin.")


@main.command()
@inp
@outp
@click.option('--style', '-s', type=str, help=style_help, default='')
@click.option('-r', '--replace', flag_value=True, help='Replace existing styles')
@click.option('--inline/--no-inline', '-l/ ', **csskwargs)
def style(layer, output, **kwargs):
    """Add or inline the CSS styles of an SVG"""
    result = _style.add_style(layer, kwargs['style'], kwargs['replace'])
    if kwargs['inline']:
        result = _style.inline(result)
    click.echo(result.encode('utf-8'), file=output)


@main.command()
@inp
@outp
@click.option('-f', '--scale', type=int)
def scale(layer, output, **kwargs):
    '''Scale all coordinates in an SVG by a factor'''
    click.echo(_style.rescale(layer, factor=kwargs['scale']).encode('utf-8'), file=output)


crs_help = ('Specify a map projection. '
                'Accepts either an EPSG code (e.g. epsg:4456), '
                'a proj4 string, '
                'a file containing a proj4, '
                '"utm" (use local UTM), '
                '"file" (use existing), '
                '"local" (generate a local projection).')


@main.command()
@click.argument('layer', default=sys.stdin, type=click.Path(exists=True))
@click.option('-j', '--crs', type=str, metavar='KEYWORD', default=None, help=crs_help)
def bounds(layer, crs):
    '''Return the bounds for a given layer.'''
    a = projection.layer_bounds(layer, crs)
    click.echo('{} {} {} {}'.format(*a), file=sys.stdout)


# Draw
@main.command()
@click.argument('layer', nargs=-1, type=str, required=True)
@click.option('-o', '--output', default=sys.stdout, type=click.File('wb'), help="Defaults to stdout.")
@click.option('-b', '--bounds', nargs=4, type=float, metavar="minx, miny, maxx, maxy", help='In the same coordinate system as the input layers', default=None)
@click.option('-c', '--style', type=str, metavar='CSS', help="CSS file or string")
@click.option('-f', '--scale', type=int, default=None, help='Scale for the map (units are divided by this number)')
@click.option('-p', '--padding', type=int, default=None, required=None, help='Buffer the map (in projection units)')
@click.option('-i', '--id-field', type=str, metavar='FIELD', help='Geodata field to use as ID')
@click.option('-a', '--class-fields', type=str, default='', metavar='FIELDS', help='Geodata fields to use as class (comma-separated)')
@click.option('-j', '--crs', default='local', metavar='KEYWORD', type=str, help=crs_help)
@click.option('-s', '--simplify', **simplifykwargs)
@click.option('--clip/--no-clip', ' /-n', **clipkwargs)
@click.option('--viewbox/--no-viewbox',  ' /-x', default=True, help='Draw SVG with or without a ViewBox. Drawing without may improve compatibility.')
@click.option('--inline/--no-inline', '-l/ ', **csskwargs)
@click.option('-q', '--quiet', default=False, flag_value=True, help='Ignore warnings.')
@click.option('-v', '--verbose', default=False, flag_value=True, help='Talk a lot.')
def draw(layer, output, **kwargs):
    '''Draw SVGs from input geodata'''

    log = logging.getLogger('svgis')

    if kwargs.pop('verbose', None):
        log.handlers[0].setLevel(logging.INFO)
        log.setLevel(logging.INFO)

    if kwargs.pop('quiet', None):
        log.handlers[0].setLevel(logging.ERROR)
        log.setLevel(logging.ERROR)

    click.echo(svgis.map(layer, **kwargs).encode('utf-8'), file=output)
    log.info('writing %s', output.name)


# Proj
@main.command()
@click.argument('bounds', nargs=4, type=float, metavar="minx, miny, maxx, maxy", default=None)
@click.option('-m', '--method', default='local', type=click.Choice(('utm', 'local')), help='Defaults to local.')
def project(bounds, method):
    '''Get a local Transverse Mercator or UTM projection for a bounding box. Expects WGS84 coordinates.'''
    click.echo(projection.generatecrs(*bounds, proj_method=method).encode('utf-8'))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of svgis.
# https://github.com/fitnr/svgis

# Licensed under the GNU General Public License v3 (GPLv3) license:
# http://opensource.org/licenses/GPL-3.0
# Copyright (c) 2016, Neil Freeman <contact@fakeisthenewreal.org>

import os.path
from fiona import transform
import fiona.crs
import fionautil.layer
import pyproj
import utm
from . import convert

'''
Helpers for dealing with projections
'''


def tm_proj4(x0, y0, y1):
    return ('+proj=lcc +lon_0={x0} +lat_1={y1} +lat_2={y0} +lat_0={y1}'
            '+x_0=0 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0'
            '+units=m +no_defs').format(x0=x0, y0=y0, y1=y1)


def utm_proj4(x, y):
    '''Generate the proj4 string for a given (lon, lat) coordinate.'''
    try:
        _, _, zonenumber, zoneletter = utm.from_latlon(y, x)
        return zonetoproj4(zonenumber, zoneletter)

    except utm.error.OutOfRangeError as e:
        raise ValueError(e)


def zonetoproj4(zonenumber, zoneletter):
    if zoneletter in 'ZYXWVUTSRQPN':
        hemisphere = 'north'

    elif zoneletter in 'MLKJHGFEDCBA':
        hemisphere = 'south'

    return '+proj=utm +zone={} +{} +datum=WGS84 +units=m +no_defs'.format(zonenumber, hemisphere)


def generatecrs(minx, miny, maxx, maxy, proj_method=None):
    '''Choose a projection, either the local UTM zone or
    create a custom transverse mercator.

    Returns:
        proj4 string
    '''
    if proj_method == 'utm':
        midx = (minx + maxx) / 2
        midy = (miny + maxy) / 2

        return utm_proj4(midx, midy)

    else:
        # Create a custom TM projection
        x0 = (float(minx) + float(maxx)) // 2

        return tm_proj4(x0, miny, maxy)


def choosecrs(in_crs, bounds, proj_method=None):
    '''Choose a projection. If the layer is projected, use that.
    Otherwise, create use a passed projection or create a custom transverse mercator.

    Args:
        in_crs (dict): A fiona-type proj4 dictionary
        bounds (tuple): (minx, miny, maxx, maxy)
        proj_method (string): either 'utm' or 'local'

    Returns:
        fiona-type proj4 dict.
    '''
    if (proj_method is None or proj_method == 'file') and not pyproj.Proj(**in_crs).is_latlong():
        # it's projected already, so noop.
        return in_crs

    else:
        return fiona.crs.from_string(generatecrs(*bounds, proj_method=proj_method))


def transform_bounds(in_crs, out_crs, bounds):
    '''
    Project a bounding box, taking care to not slice off the sides.
    '''
    if in_crs is None:
        raise TypeError('Need input CRS, not None')

    if out_crs is None:
        raise TypeError('Need output CRS, not None')

    ring = convert.bounds_to_ring(*bounds)
    xs, ys = list(zip(*ring))

    xbounds, ybounds = transform.transform(in_crs, out_crs, xs, ys)

    return min(xbounds), min(ybounds), max(xbounds), max(ybounds)


def layer_bounds(layer, crs=None):
    '''
    Get the bounds of a layer, optionally transforming them into a given CRS (or local or utm).

    Args:
        layer (str): path to a geodata file.
        crs (str): (optional) Any of the crs specs typically accepted
                   by SVGIS (An EPSG code, a proj4 string, a file containing a proj4, 'local' or 'utm')

    Returns:
        tuple
    '''
    meta = fionautil.layer.meta_complete(layer)

    if crs:
        _, crs = pick(crs)
        result = transform_bounds(meta['crs'], crs, meta['bounds'])
    else:
        result = meta['bounds']

    return result


def pick(project):
    '''
        Pick a projection or projection method to use.

    Returns:
        tuple: (projection method, crs)
    '''
    proj_method, out_crs = None, None

    if project is None:
        project = 'local'

    if project.lower() in ('local', 'utm'):
        proj_method = project.lower()

    elif os.path.exists(project):
        # Is a file
        with open(project) as f:
            out_crs = fiona.crs.from_string(f.read())

    elif project[:5].lower() == 'epsg:':
        # Is an epsg code
        _, epsg = project.split(':')
        out_crs = fiona.crs.from_epsg(int(epsg))

    else:
        # Assume it's a proj4 string.
        # fiona.crs.from_string returns {} if it isn't.
        out_crs = fiona.crs.from_string(project)

    return proj_method, out_crs
